"""
Classes for working with surface (`.obj`) files.
"""

from abc import ABC

from civet.abstract_data import DataSource, DataFile
from civet.xfm import XfmTransformationMixin


class HemisphereMixin(DataSource, ABC):
    """
    Provides a subclass with XFM-related helper functions for `.obj` files.
    """
    preferred_suffix = '.obj'

    def flip_x(self) -> 'HemisphereMixin':
        """
        Flip this surface along the *x*-axis.
        """
        return _XfmObjTransformation(self, '-scales', -1, 1, 1)

    def translate_x(self, x: float) -> 'HemisphereMixin':
        """
        Translate this surface along the *x*-axis.
        """
        return _XfmObjTransformation(self, '-translation', x, 0, 0)

    def slide_left(self) -> 'HemisphereMixin':
        """
        Translate this surface 25 units to the left.
        """
        return self.translate_x(-25)

    def slide_right(self) -> 'HemisphereMixin':
        """
        Translate this surface 25 units to the right.
        """
        return self.translate_x(25)


class _XfmObjTransformation(HemisphereMixin, XfmTransformationMixin):
    transform_program = 'transform_objects'


class ObjFile(DataFile, HemisphereMixin):
    """
    Wraps a MNI `.obj` file, which is a polygonal mesh representing
    the surface of a brain hemisphere.

    Examples
    --------

    ```python
    from civet import MNI_DATAPATH, ObjFile

    starting_model = ObjFile(MNI_DATAPATH / 'surface-extraction' / 'white_model_320.obj')
    starting_model.flip_x().slide_right().save('./output.obj')
    ```
    """
    pass



"""
Wrappers for data paths and programs which process data.
"""
import abc
import shutil
from dataclasses import dataclass
from os import path, PathLike
from tempfile import NamedTemporaryFile


class DataSource(abc.ABC):
    """
    A `DataSource` is something which can provide data, such as an
    input file, input directory, a command which processes input, or
    a chain of commands. In the latter case, the commands are executed
    "lazily" and will only actually run when `DataSource.save` is called.
    Intermediate files are written to temporary files and then removed.
    """

    preferred_suffix: str = ''
    """
    Preferred output file suffix.
    """

    @abc.abstractmethod
    def save(self, output: PathLike | str) -> None:
        """
        Resolve this data to a given path by executing the
        represented operations.
        """
        ...


@dataclass(frozen=True)
class DataFile(DataSource):
    """
    A wrapper around a file path.
    """
    input: str | PathLike

    def __post_init__(self):
        if not path.exists(self.input):
            raise FileNotFoundError(f'Input does not exist: {self.input}')

    def save(self, output: PathLike | str) -> None:
        """
        Copy the file to a given destination.
        """
        shutil.copyfile(self.input, output)


@dataclass(frozen=True)
class DataOperation(DataSource, abc.ABC):
    """
    A recursive data structure representing an operation to perform on input data.
    """

    input: DataSource

    @abc.abstractmethod
    def run(self, input: str | PathLike, output: str | PathLike):
        ...

    def save(self, output: str | PathLike):
        """
        Perform the operation on its input data.
        """
        with NamedTemporaryFile(suffix=self.input.preferred_suffix) as real_input:
            self.input.save(real_input.name)
            self.run(real_input.name, output)



"""
Classes for working with volume (`.mnc`) files.
"""

from os import PathLike
from abc import ABC
import subprocess as sp
from civet.abstract_data import DataSource, DataOperation, DataFile


class MincMixin(DataSource, ABC):
    """
    Provides a subclass with MINC volume related functions.
    """
    preferred_suffix = '.mnc'

    def reshape_bbox(self) -> 'MincMixin':
        """
        Runs

        ```shell
        mincreshape -quiet -clobber $(mincbbox -mincreshape in.mnc) in.mnc out.mnc
        ```
        """
        return _MincBBoxReshapeOperation(self)


class _MincBBoxReshapeOperation(DataOperation, MincMixin):
    def run(self, input: str | PathLike, output: str | PathLike):
        cmd = [
            'mincreshape', '-quiet', '-clobber',
            *self._mincbbox(input), input, output
        ]
        sp.run(cmd, check=True)

    def _mincbbox(self, input: str | PathLike) -> list[str]:
        result = sp.check_output(['mincbbox', '-mincreshape', input])
        decoded = result.decode(encoding='utf-8')
        return decoded.split(' ')


class MincFile(DataFile, MincMixin):
    """
    Wraps a MINC (`.mnc`) file.

    Examples
    --------

    ```python
    from civet import MincFile

    MincFile('mask.mnc').reshape_bbox().save('bounded.mnc')
    ```
    """
    pass



"""
Object-oriented Python bindings for CIVET binaries such as
`transform_objects` and `mincreshape`.
"""

import civet.minc as minc
# from civet.minc import MincFile, MincMixin
# from civet.surface import ObjFile, HemisphereMixin
import civet.surface as surface
# from civet.abstract_data import DataSource, DataFile
import civet.abstract_data as abstract_data

import os
from pathlib import Path

__fallback_path = '/runtime/path/to/CIVET/quarantines/Linux-x86_64/share'
MNI_DATAPATH = Path(os.getenv('MNI_DATAPATH', __fallback_path))
"""
Location of data such as starting ellipsoids for surface extraction.
"""

__docformat__ = 'numpy'

__all__ = [
    'MNI_DATAPATH',
    'minc',
    # 'MincFile',
    # 'MincMixin',
    # 'ObjFile',
    # 'HemisphereMixin',
    'surface',
    # 'DataFile',
    # 'DataSource'
    'abstract_data'
]



import abc
from os import PathLike
import subprocess as sp
from dataclasses import dataclass
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
from typing import Literal
from civet.abstract_data import DataOperation


@dataclass(frozen=True)
class XfmTransformationMixin(DataOperation, abc.ABC):
    """
    Chains the `param2xfm` command and a `transform_objects` or `transform_volume` command.
    """

    op: str
    x: float
    y: float
    z: float

    def run(self, input: str | PathLike, output: str | PathLike):
        with self._param2xfm(self.op, self.x, self.y, self.z) as xfm:
            sp.run([self.transform_program, input, xfm, output], check=True)

    @staticmethod
    @contextmanager
    def _param2xfm(*args):
        with NamedTemporaryFile(suffix='.xfm') as f:
            sp.run(['param2xfm', '-clobber', *map(str, args), f.name], check=True)
            yield f.name

    @property
    @abc.abstractmethod
    def transform_program(self) -> Literal['transform_objects', 'transform_volume']:
        ...



from setuptools import setup
import os
import shutil

if 'CI' not in os.environ and not shutil.which('adapt_object_mesh'):
    raise Exception(
        'CIVET binaries such as "adapt_object_mesh" not found in $PATH.'
        '\nHint: consider using container images, such as '
        '"docker.io/fnndsc/mni-conda-base:civet2.1.1-python3.10.2"'
    )

with open('README.md', 'r') as f:
    readme = f.read()

setup(
    name='pycivet',
    version='0.0.1a1',
    description='CIVET bindings',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Jennings Zhang',
    author_email='Jennings.Zhang@childrens.harvard.edu',
    url='https://github.com/FNNDSC/pycivet',
    project_urls={
        'Documentation': 'https://fnndsc.github.io/pycivet/',
        'Source': 'https://github.com/FNNDSC/pycivet',
        'Tracker': 'https://github.com/FNNDSC/pycivet/issues',
    },
    license='MIT',
    packages=['civet'],
    python_requires='>=3.10.2',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.10',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ]
)
